indexing

	description: "Base class for a service."


deferred class

	SERVICE


inherit

	ANY

	WEL_ERROR
		export
			{NONE} all
		end

	SERVICE_DEBUGGING
		export
			{NONE} all
		end

	WINSVC_FUNCTIONS_EXTERNAL
		export
			{NONE} all
		end

	LPHANDLER_FUNCTION_CALLBACK
		rename
			on_callback as controller
		end

	WG_EXTERNAL_STRING_ROUTINES
		export
			{NONE} all
		end


feature {NONE} -- Initialization

	make is
		do
			create status.make_new_shared
			status.set_dwservicetype (service_win32_own_process)
			status.set_dwcontrolsaccepted (service_accept_stop)
			status.set_dwwaithint (default_wait_hint)
		end


feature -- Access

	cm_service_control_code: INTEGER is 0x0401
			-- Message code for communication with service thread.
			-- WM_USER + 1

	name: STRING is
		deferred
		ensure
			name_not_empty: Result /= Void and then not Result.is_empty
		end

	status: SERVICE_STATUS_STRUCT
			-- Current status of service.

	status_handle: INTEGER
			-- Service status handle if not zero.

	default_wait_hint: INTEGER is
		do
			Result := 5000
		end


feature -- Status

	is_status_handle_valid: BOOLEAN is
			-- Does `status_handle' refer to a valid status handle?
		do
			Result := status_handle /= 0
		end


feature -- Status change

	report_status is
			-- Report status to SCM.
		require
			have_status_handle: is_status_handle_valid
		local
			success: BOOLEAN
		do
			success := set_service_status_external (status_handle, status.item) /= 0
			if not success then
				log ("report_status failed.")
			end
		end


	set_status (new_status: INTEGER) is
			-- Change current status in `status' and report it.
		require
			have_status_handle: is_status_handle_valid
		local
			last_status: INTEGER
		do
			last_status := status.get_dwcurrentstate
			status.set_dwcurrentstate (new_status)
			-- Reset dwCheckPoint and dwWaitHint when we're running
			inspect new_status
			when service_start_pending then
				-- Assume wait hint in struct is set correctly
				if last_status = service_start_pending then
					status.set_dwcheckpoint (status.get_dwcheckpoint + 1)
				else
					status.set_dwcheckpoint (0)
				end
			when service_running, service_stopped, service_paused then
				status.set_dwcheckpoint (0)
				status.set_dwwaithint (0)
			else
				-- nothing special
			end
			report_status
		end


feature -- Main

	main (arguments: ARRAY [STRING]) is
			-- Start of service. Called in separate thread.
			-- Should not call any win32 APIs until status is set to
			-- service_running, according the the MSDN
			-- documentation. That won't work, I'm afraid.
		local
			exceptions: expanded EXCEPTIONS
		do
			create callback.make (Current)
			-- Want to use RegisterServiceCtrlHandlerEx here
			-- is in Borland C headers, not in the M$ headers I use
			status_handle := register_service_ctrl_handler_a_external (string_to_pointer (name), callback.c_dispatcher)
			if status_handle /= 0 then
				set_status (service_start_pending)
				run_thread
				debug ("service")
					log ("service.main end")
				end
				set_status (service_stopped)
			else
				debug ("service")
					log ("service.main failed")
				end
			end
		rescue
			debug ("service")
				log (exceptions.tag_name)
			end
		end


feature {SERVICE_THREAD} -- Service execution

	execute is
		require
			thread_not_void: thread /= Void
		do
			debug ("service")
				log ("service.execute begin")
			end
			set_status (service_running)
			debug ("service")
				log ("service.execute before process requests")
			end
			thread.process_requests (True)
			debug ("service")
				log ("service.execute end")
			end
		rescue
			debug ("service")
				log ("service.execute rescue")
			end
		end

	interrogate is
		do
			debug ("service")
				log ("service.interrogate")
			end
			report_status
		end

	shutdown is
		require
			thread_not_void: thread /= Void
		do
			debug ("service")
				log ("service.shutdown")
			end
			set_status (service_stop_pending)
			thread.terminate
		end

	stop is
		require
			thread_not_void: thread /= Void
		do
			debug ("service")
				log ("service.stop begin")
			end
			set_status (service_stop_pending)
			thread.terminate
			debug ("service")
				log ("service.stop end")
			end
		end


feature {NONE} -- Service thread

	run_thread is
			-- Routine may never raise an exception.
		local
			retried: BOOLEAN
		do
			if not retried then
				debug ("service")
					log ("service.start_thread before create thread")
				end
				create thread.make (Current)
				thread.launch
				debug ("service")
					log ("service.start_thread before join")
				end
				thread.join
				debug ("service")
					log ("service.start_thread after join")
				end
				thread := Void
			end
		rescue
			retried := True
			retry
		end


feature {SERVICE_THREAD} -- Service thread

	thread: SERVICE_THREAD
			-- Thread that does the actual work done by this service.


feature {NONE} -- Controller callbacks

	callback: LPHANDLER_FUNCTION_DISPATCHER
			-- Callback for `controller'.

	controller (dwcontrol: INTEGER) is
			-- Controller handler function, called by NT in main process
			-- thread. Should communication `dwcontrol' to the service
			-- thread.
		local
			success: BOOLEAN
			error: INTEGER
		do
-- 			debug ("service")
-- 				log ("controller: " + dwcontrol.out)
-- 			end
			if thread /= Void then
				success := cwin_post_thread_message (thread.thread_id, cm_service_control_code, dwcontrol, 0)
				if not success then
					error := cwin_get_last_error
					debug ("service")
						log ("sending msg to thread failed: " + error.out)
					end
				end
			end
-- 			debug ("service")
-- 				log ("controller end")
-- 			end
		end


feature {NONE} -- Windows API

	cwin_get_last_error: INTEGER is
		external
			"C [macro <windows.h>]"
		alias
			"GetLastError()"
		end

	cwin_post_thread_message (thread_id: INTEGER; msg, wparam, lparam: INTEGER): BOOLEAN is
			-- The PostThreadMessage function places (posts) a message in
			-- the message queue of the specified thread and then returns
			-- without waiting for the thread to process the message.
		external
			"C [macro %"wel.h%"] (DWORD, UINT, WPARAM, LPARAM): EIF_BOOLEAN"
		alias
			"PostThreadMessage"
		end


feature {NONE} -- Service constants

	service_win32_own_process: INTEGER is 0x00000010
	service_win32_share_process: INTEGER is 0x00000020

	service_accept_stop: INTEGER is 0x00000001

	service_stopped: INTEGER is 1
	service_start_pending: INTEGER is 2
	service_stop_pending: INTEGER is 3
	service_running: INTEGER is 4
	service_continue_pending: INTEGER is 5
	service_pause_pending: INTEGER is 6
	service_paused: INTEGER is 7


invariant

	status_not_void: status /= Void

end
